@@ -2,24 +2,23 @@ from __future__ import print_function, division, absolute_import
 
 from functools import partial
 import os
-import ssl
 import sys
 import threading
 
 import pytest
 
-from tornado import gen, ioloop, queues
+from tornado import gen, ioloop, locks, queues
 from tornado.concurrent import Future
 
-from distributed.core import pingpong
 from distributed.metrics import time
 from distributed.utils import get_ip, get_ipv6
-from distributed.utils_test import (slow, loop, gen_test, gen_cluster,
-                                    requires_ipv6, has_ipv6, get_cert,
-                                    get_server_ssl_context, get_client_ssl_context)
+from distributed.utils_test import (gen_test, requires_ipv6, has_ipv6,
+                                    get_cert, get_server_ssl_context,
+                                    get_client_ssl_context)
+from distributed.utils_test import loop  # flake8: noqa
 
-from distributed.protocol import (loads, dumps,
-                                  to_serialize, Serialized, serialize, deserialize)
+from distributed.protocol import (to_serialize, Serialized, serialize,
+                                  deserialize)
 
 from distributed.comm import (tcp, inproc, connect, listen, CommClosedError,
                               parse_address, parse_host_port,
@@ -40,7 +39,8 @@ cert_subject = (
     (('localityName', 'Dask-distributed'),),
     (('organizationName', 'Dask'),),
     (('commonName', 'localhost'),)
-    )
+)
+
 
 def check_tls_extra(info):
     assert isinstance(info, dict)
@@ -58,30 +58,34 @@ tls_kwargs = dict(listen_args={'ssl_context': get_server_ssl_context()},
 
 
 @gen.coroutine
-def get_comm_pair(listen_addr, listen_args=None, connect_args=None):
+def get_comm_pair(listen_addr, listen_args=None, connect_args=None,
+                  **kwargs):
     q = queues.Queue()
 
     def handle_comm(comm):
         q.put(comm)
 
     listener = listen(listen_addr, handle_comm,
-                      connection_args=listen_args)
+                      connection_args=listen_args, **kwargs)
     listener.start()
 
     comm = yield connect(listener.contact_address,
-                         connection_args=connect_args)
+                         connection_args=connect_args, **kwargs)
     serv_comm = yield q.get()
     raise gen.Return((comm, serv_comm))
 
 
-def get_tcp_comm_pair():
-    return get_comm_pair('tcp://')
+def get_tcp_comm_pair(**kwargs):
+    return get_comm_pair('tcp://', **kwargs)
 
-def get_tls_comm_pair():
-    return get_comm_pair('tls://', **tls_kwargs)
 
-def get_inproc_comm_pair():
-    return get_comm_pair('inproc://')
+def get_tls_comm_pair(**kwargs):
+    kwargs.update(tls_kwargs)
+    return get_comm_pair('tls://', **kwargs)
+
+
+def get_inproc_comm_pair(**kwargs):
+    return get_comm_pair('inproc://', **kwargs)
 
 
 @gen.coroutine
@@ -322,6 +326,8 @@ def check_inproc_specific(run_client):
             msg = yield comm.read()
         assert msg == {'op': 'pong', 'data': key}
         l.append(key)
+        with pytest.raises(CommClosedError):
+            yield comm.read()
         yield comm.close()
 
     client_communicate = partial(run_client, client_communicate)
@@ -341,6 +347,7 @@ def check_inproc_specific(run_client):
 def run_coro(func, *args, **kwargs):
     return func(*args, **kwargs)
 
+
 def run_coro_in_thread(func, *args, **kwargs):
     fut = Future()
     main_loop = ioloop.IOLoop.current()
@@ -350,7 +357,7 @@ def run_coro_in_thread(func, *args, **kwargs):
         try:
             res = thread_loop.run_sync(partial(func, *args, **kwargs),
                                        timeout=10)
-        except:
+        except Exception:
             main_loop.add_callback(fut.set_exc_info, sys.exc_info())
         else:
             main_loop.add_callback(fut.set_result, res)
@@ -366,6 +373,7 @@ def run_coro_in_thread(func, *args, **kwargs):
 def test_inproc_specific_same_thread():
     yield check_inproc_specific(run_coro)
 
+
 @gen_test()
 def test_inproc_specific_different_threads():
     yield check_inproc_specific(run_coro_in_thread)
@@ -460,8 +468,10 @@ def tcp_eq(expected_host, expected_port=None):
 
     return checker
 
+
 tls_eq = tcp_eq
 
+
 def inproc_check():
     expected_ip = get_ip()
     expected_pid = os.getpid()
@@ -489,6 +499,7 @@ def test_default_client_server_ipv4():
     yield check_client_server(':3203',
                               tcp_eq('0.0.0.0', 3203), tcp_eq(EXTERNAL_IP4, 3203))
 
+
 @requires_ipv6
 @gen_test()
 def test_default_client_server_ipv6():
@@ -497,6 +508,7 @@ def test_default_client_server_ipv6():
     yield check_client_server('[::]', tcp_eq('::'), tcp_eq(EXTERNAL_IP6))
     yield check_client_server('[::]:3212', tcp_eq('::', 3212), tcp_eq(EXTERNAL_IP6, 3212))
 
+
 @gen_test()
 def test_tcp_client_server_ipv4():
     yield check_client_server('tcp://127.0.0.1', tcp_eq('127.0.0.1'))
@@ -529,6 +541,7 @@ def test_tls_client_server_ipv4():
     yield check_client_server('tls://', tls_eq('0.0.0.0'),
                               tls_eq(EXTERNAL_IP4), **tls_kwargs)
 
+
 @requires_ipv6
 @gen_test()
 def test_tls_client_server_ipv6():
@@ -627,10 +640,12 @@ def check_comm_closed_implicit(addr, delay=None, listen_args=None,
 def test_tcp_comm_closed_implicit():
     yield check_comm_closed_implicit('tcp://127.0.0.1')
 
+
 @gen_test()
 def test_tls_comm_closed_implicit():
     yield check_comm_closed_implicit('tls://127.0.0.1', **tls_kwargs)
 
+
 @gen_test()
 def test_inproc_comm_closed_implicit():
     yield check_comm_closed_implicit(inproc.new_address())
@@ -638,43 +653,43 @@ def test_inproc_comm_closed_implicit():
 
 @gen.coroutine
 def check_comm_closed_explicit(addr, listen_args=None, connect_args=None):
-    @gen.coroutine
-    def handle_comm(comm):
-        # Wait
-        try:
-            yield comm.read()
-        except CommClosedError:
-            pass
-
-    listener = listen(addr, handle_comm, connection_args=listen_args)
-    listener.start()
-    contact_addr = listener.contact_address
-
-    comm = yield connect(contact_addr, connection_args=connect_args)
-    comm.close()
+    a, b = yield get_comm_pair(addr, listen_args=listen_args, connect_args=connect_args)
+    a_read = a.read()
+    b_read = b.read()
+    yield a.close()
+    # In-flight reads should abort with CommClosedError
     with pytest.raises(CommClosedError):
-        yield comm.write({})
-
-    comm = yield connect(contact_addr, connection_args=connect_args)
-    comm.close()
+        yield a_read
     with pytest.raises(CommClosedError):
-        yield comm.read()
-
-    yield gen.moment
+        yield b_read
+    # New reads as well
+    with pytest.raises(CommClosedError):
+        yield a.read()
+    with pytest.raises(CommClosedError):
+        yield b.read()
+    # And writes
+    with pytest.raises(CommClosedError):
+        yield a.write({})
+    with pytest.raises(CommClosedError):
+        yield b.write({})
+    yield b.close()
 
 
 @gen_test()
 def test_tcp_comm_closed_explicit():
     yield check_comm_closed_explicit('tcp://127.0.0.1')
 
+
 @gen_test()
 def test_tls_comm_closed_explicit():
     yield check_comm_closed_explicit('tls://127.0.0.1', **tls_kwargs)
 
+
 @gen_test()
 def test_inproc_comm_closed_explicit():
     yield check_comm_closed_explicit(inproc.new_address())
 
+
 @gen_test()
 def test_inproc_comm_closed_explicit_2():
     listener_errors = []
@@ -739,6 +754,7 @@ def check_connect_timeout(addr):
 def test_tcp_connect_timeout():
     yield check_connect_timeout('tcp://127.0.0.1:44444')
 
+
 @gen_test()
 def test_inproc_connect_timeout():
     yield check_connect_timeout(inproc.new_address())
@@ -770,6 +786,7 @@ def test_tcp_many_listeners():
     check_many_listeners('tcp://0.0.0.0')
     check_many_listeners('tcp://')
 
+
 @gen_test()
 def test_inproc_many_listeners():
     check_many_listeners('inproc://')
@@ -793,31 +810,36 @@ def check_listener_deserialize(addr, deserialize, in_value, check_out):
         comm = yield connect(listener.contact_address)
 
     yield comm.write(in_value)
-    yield comm.close()
 
     out_value = yield q.get()
     check_out(out_value)
+    yield comm.close()
+
 
 @gen.coroutine
 def check_connector_deserialize(addr, deserialize, in_value, check_out):
-    q = queues.Queue()
+    done = locks.Event()
 
     @gen.coroutine
     def handle_comm(comm):
-        msg = yield q.get()
-        yield comm.write(msg)
+        yield comm.write(in_value)
+        yield done.wait()
         yield comm.close()
 
     with listen(addr, handle_comm) as listener:
         comm = yield connect(listener.contact_address, deserialize=deserialize)
 
-    q.put_nowait(in_value)
     out_value = yield comm.read()
+    done.set()
     yield comm.close()
     check_out(out_value)
 
+
 @gen.coroutine
 def check_deserialize(addr):
+    """
+    Check the "deserialize" flag on connect() and listen().
+    """
     # Test with Serialize and Serialized objects
 
     msg = {'op': 'update',
@@ -862,33 +884,103 @@ def check_deserialize(addr):
     yield check_listener_deserialize(addr, True, msg, check_out_true)
     yield check_connector_deserialize(addr, True, msg, check_out_true)
 
-    # Test with a long bytestring
+    # Test with long bytestrings, large enough to be transferred
+    # as a separate payload
+
+    _uncompressible = os.urandom(1024 ** 2) * 4  # end size: 8 MB
 
     msg = {'op': 'update',
-           'x': b'abc',
-           'y': b'def\n' * (3 * 1024 ** 2),  # end size: 12 MB
+           'x': _uncompressible,
+           'to_ser': [to_serialize(_uncompressible)],
+           'ser': Serialized(*serialize(_uncompressible)),
            }
     msg_orig = msg.copy()
 
-    def check_out(out_value):
-        assert out_value == msg_orig
+    def check_out(deserialize_flag, out_value):
+        # Check output with deserialize=False
+        assert sorted(out_value) == sorted(msg_orig)
+        out_value = out_value.copy()  # in case transport passed the object as-is
+        to_ser = out_value.pop('to_ser')
+        ser = out_value.pop('ser')
+        expected_msg = msg_orig.copy()
+        del expected_msg['ser']
+        del expected_msg['to_ser']
+        assert out_value == expected_msg
 
-    yield check_listener_deserialize(addr, False, msg, check_out)
-    yield check_connector_deserialize(addr, False, msg, check_out)
+        if deserialize_flag:
+            assert isinstance(ser, (bytes, bytearray))
+            assert bytes(ser) == _uncompressible
+        else:
+            assert isinstance(ser, Serialized)
+            assert deserialize(ser.header, ser.frames) == _uncompressible
+            assert isinstance(to_ser, list)
+            to_ser, = to_ser
+            # The to_serialize() value could have been actually serialized
+            # or not (it's a transport-specific optimization)
+            if isinstance(to_ser, Serialized):
+                assert deserialize(to_ser.header, to_ser.frames) == _uncompressible
+            else:
+                assert to_ser == to_serialize(_uncompressible)
 
-    yield check_listener_deserialize(addr, True, msg, check_out)
-    yield check_connector_deserialize(addr, True, msg, check_out)
+    yield check_listener_deserialize(addr, False, msg, partial(check_out, False))
+    yield check_connector_deserialize(addr, False, msg, partial(check_out, False))
+
+    yield check_listener_deserialize(addr, True, msg, partial(check_out, True))
+    yield check_connector_deserialize(addr, True, msg, partial(check_out, True))
 
 
 @gen_test()
 def test_tcp_deserialize():
     yield check_deserialize('tcp://')
 
+
 @gen_test()
 def test_inproc_deserialize():
     yield check_deserialize('inproc://')
 
 
+@gen.coroutine
+def check_deserialize_roundtrip(addr):
+    """
+    Sanity check round-tripping with "deserialize" on and off.
+    """
+    # Test with long bytestrings, large enough to be transferred
+    # as a separate payload
+    _uncompressible = os.urandom(1024 ** 2) * 4  # end size: 4 MB
+
+    msg = {'op': 'update',
+           'x': _uncompressible,
+           'to_ser': [to_serialize(_uncompressible)],
+           'ser': Serialized(*serialize(_uncompressible)),
+           }
+
+    for deserialize in (True, False):
+        a, b = yield get_comm_pair(addr, deserialize=deserialize)
+        yield a.write(msg)
+        got = yield b.read()
+        yield b.write(got)
+        got = yield a.read()
+
+        assert sorted(got) == sorted(msg)
+        for k in ('op', 'x'):
+            assert got[k] == msg[k]
+        if deserialize:
+            assert isinstance(got['to_ser'][0], (bytes, bytearray))
+            assert isinstance(got['ser'], (bytes, bytearray))
+        else:
+            assert isinstance(got['to_ser'][0], (to_serialize, Serialized))
+            assert isinstance(got['ser'], Serialized)
+
+
+@gen_test()
+def test_inproc_deserialize_roundtrip():
+    yield check_deserialize_roundtrip('inproc://')
+
+
+@gen_test()
+def test_tcp_deserialize_roundtrip():
+    yield check_deserialize_roundtrip('tcp://')
+
 
 def _raise_eoferror():
     raise EOFError
@@ -915,6 +1007,7 @@ def check_deserialize_eoferror(addr):
         with pytest.raises(CommClosedError):
             yield comm.read()
 
+
 @gen_test()
 def test_tcp_deserialize_eoferror():
     yield check_deserialize_eoferror('tcp://')
@@ -933,6 +1026,7 @@ def check_repr(a, b):
     yield b.close()
     assert 'closed' in repr(b)
 
+
 @gen_test()
 def test_tcp_repr():
     a, b = yield get_tcp_comm_pair()
@@ -940,6 +1034,7 @@ def test_tcp_repr():
     assert b.local_address in repr(a)
     yield check_repr(a, b)
 
+
 @gen_test()
 def test_tls_repr():
     a, b = yield get_tls_comm_pair()
@@ -947,6 +1042,7 @@ def test_tls_repr():
     assert b.local_address in repr(a)
     yield check_repr(a, b)
 
+
 @gen_test()
 def test_inproc_repr():
     a, b = yield get_inproc_comm_pair()
@@ -962,16 +1058,19 @@ def check_addresses(a, b):
     a.abort()
     b.abort()
 
+
 @gen_test()
 def test_tcp_adresses():
     a, b = yield get_tcp_comm_pair()
     yield check_addresses(a, b)
 
+
 @gen_test()
 def test_tls_adresses():
     a, b = yield get_tls_comm_pair()
     yield check_addresses(a, b)
 
+
 @gen_test()
 def test_inproc_adresses():
     a, b = yield get_inproc_comm_pair()
