@@ -6,21 +6,24 @@ import logging
 import six
 import traceback
 import uuid
+import weakref
 
 from six import string_types
 
 from toolz import assoc
 
 from tornado import gen
-from tornado.ioloop import IOLoop, PeriodicCallback
+from tornado.ioloop import IOLoop
 from tornado.locks import Event
 
 from .comm import (connect, listen, CommClosedError,
                    normalize_address,
                    unparse_host_port, get_address_host_port)
+from .config import config
 from .metrics import time
 from .system_monitor import SystemMonitor
-from .utils import get_traceback, truncate_exception, ignoring, shutting_down
+from .utils import (get_traceback, truncate_exception, ignoring, shutting_down,
+                    PeriodicCallback)
 from . import protocol
 
 
@@ -75,7 +78,8 @@ class Server(object):
     So in the example above the following would be good messages.
 
     *  ``{'op': 'ping'}``
-    *  ``{'op': 'add': 'x': 10, 'y': 20}``
+    *  ``{'op': 'add', 'x': 10, 'y': 20}``
+
     """
     default_ip = ''
     default_port = 0
@@ -83,7 +87,7 @@ class Server(object):
     def __init__(self, handlers, connection_limit=512, deserialize=True,
                  io_loop=None):
         self.handlers = assoc(handlers, 'identity', self.identity)
-        self.id = type(self).__name__ + '-' + str(uuid.uuid1())
+        self.id = type(self).__name__ + '-' + str(uuid.uuid4())
         self._address = None
         self._listen_address = None
         self._port = None
@@ -94,6 +98,7 @@ class Server(object):
         self.digests = None
         self.events = None
         self.event_counts = None
+        self._ongoing_coroutines = weakref.WeakSet()
 
         self.listener = None
         self.io_loop = io_loop or IOLoop.current()
@@ -108,16 +113,29 @@ class Server(object):
         self.events = defaultdict(lambda: deque(maxlen=10000))
         self.event_counts = defaultdict(lambda: 0)
 
-        pc = PeriodicCallback(self.monitor.update, 500, io_loop=self.io_loop)
-        self.io_loop.add_callback(pc.start)
-        if self.digests is not None:
-            self._last_tick = time()
-            self._tick_pc = PeriodicCallback(self._measure_tick, 20,
-                                             io_loop=self.io_loop)
-            self.io_loop.add_callback(self._tick_pc.start)
+        self.periodic_callbacks = dict()
+
+        pc = PeriodicCallback(self.monitor.update, 500)
+        self.periodic_callbacks['monitor'] = pc
+
+        self._last_tick = time()
+        pc = PeriodicCallback(self._measure_tick, config.get('tick-time', 20))
+        self.periodic_callbacks['tick'] = pc
 
         self.__stopped = False
 
+    def start_periodic_callbacks(self):
+        """ Start Periodic Callbacks consistently
+
+        This starts all PeriodicCallbacks stored in self.periodic_callbacks if
+        they are not yet running.  It does this safely on the IOLoop.
+        """
+        def start_pcs():
+            for pc in self.periodic_callbacks.values():
+                if not pc.is_running():
+                    pc.start()
+        self.loop.add_callback(start_pcs)
+
     def stop(self):
         if not self.__stopped:
             self.__stopped = True
@@ -134,13 +152,15 @@ class Server(object):
     def _measure_tick(self):
         now = time()
         diff = now - self._last_tick
-        self.digests['tick-duration'].add(diff)
         self._last_tick = now
-        if diff > 1:
-            logger.warn("Event loop was unresponsive for %.1fs.  "
-                        "This is often caused by long-running GIL-holding "
-                        "functions.  This can cause timeouts and instability.",
-                        diff)
+        if diff > config.get('tick-maximum-delay', 1000) / 1000:
+            logger.warning("Event loop was unresponsive for %.2fs.  "
+                           "This is often caused by long-running GIL-holding "
+                           "functions or moving large chunks of data. "
+                           "This can cause timeouts and instability.",
+                           diff)
+        if self.digests is not None:
+            self.digests['tick-duration'].add(diff)
 
     def log_event(self, name, msg):
         msg['time'] = time()
@@ -233,8 +253,8 @@ class Server(object):
                 except EnvironmentError as e:
                     if not shutting_down():
                         logger.debug("Lost connection to %r while reading message: %s."
-                                    " Last operation: %s",
-                                    address, e, op)
+                                     " Last operation: %s",
+                                     address, e, op)
                     break
                 except Exception as e:
                     logger.exception(e)
@@ -263,6 +283,7 @@ class Server(object):
                     try:
                         result = handler(comm, **msg)
                         if type(result) is gen.Future:
+                            self._ongoing_coroutines.add(result)
                             result = yield result
                     except CommClosedError as e:
                         logger.warning("Lost connection to %r: %s", address, e)
@@ -274,8 +295,8 @@ class Server(object):
                     try:
                         yield comm.write(result)
                     except EnvironmentError as e:
-                        logger.warning("Lost connection to %r while sending result for op %r: %s",
-                                       address, op, e)
+                        logger.debug("Lost connection to %r while sending result for op %r: %s",
+                                     address, op, e)
                         break
                 msg = result = None
                 if close_desired:
@@ -292,6 +313,19 @@ class Server(object):
                     logger.error("Failed while closing connection to %r: %s",
                                  address, e)
 
+    @gen.coroutine
+    def close(self):
+        self.listener.stop()
+        for comm in self._comms:
+            comm.close()
+        for cb in self._ongoing_coroutines:
+            cb.cancel()
+        for i in range(10):
+            if all(cb.cancelled() for c in self._ongoing_coroutines):
+                break
+            else:
+                yield gen.sleep(0.01)
+
 
 def pingpong(comm):
     return b'pong'
@@ -355,7 +389,7 @@ class rpc(object):
 
     >>> remote.close_comms()  # doctest: +SKIP
     """
-    active = 0
+    active = weakref.WeakSet()
     comms = ()
     address = None
 
@@ -367,7 +401,7 @@ class rpc(object):
         self.status = 'running'
         self.deserialize = deserialize
         self.connection_args = connection_args
-        rpc.active += 1
+        rpc.active.add(self)
 
     @gen.coroutine
     def live_comm(self):
@@ -438,7 +472,7 @@ class rpc(object):
 
     def close_rpc(self):
         if self.status != 'closed':
-            rpc.active -= 1
+            rpc.active.discard(self)
         self.status = 'closed'
         self.close_comms()
 
@@ -450,7 +484,7 @@ class rpc(object):
 
     def __del__(self):
         if self.status != 'closed':
-            rpc.active -= 1
+            rpc.active.discard(self)
             self.status = 'closed'
             still_open = [comm for comm in self.comms if not comm.closed()]
             if still_open:
@@ -469,6 +503,7 @@ class PooledRPCCall(object):
     See Also:
         ConnectionPool
     """
+
     def __init__(self, addr, pool):
         self.addr = addr
         self.pool = pool
@@ -532,6 +567,7 @@ class ConnectionPool(object):
     deserialize: bool
         Whether or not to deserialize data by default or pass it through
     """
+
     def __init__(self, limit=512, deserialize=True, connection_args=None):
         self.open = 0          # Total number of open comms
         self.active = 0        # Number of comms currently in use
@@ -544,11 +580,9 @@ class ConnectionPool(object):
         self.connection_args = connection_args
         self.event = Event()
 
-    def __str__(self):
+    def __repr__(self):
         return "<ConnectionPool: open=%d, active=%d>" % (self.open,
-                self.active)
-
-    __repr__ = __str__
+                                                         self.active)
 
     def __call__(self, addr=None, ip=None, port=None):
         """ Cached rpc objects """
